# Open a console
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import threading

from xen.xend.xenstore.xstransact import xstransact
from xen.xend.xenstore.xswatch import xswatch

from log import log
from VMDefaults import VMDefaults
import msg
import util

_reset_console = False


def get_domid_from_uuid(uuid):
    vm_path = '/vm/%s' % uuid
    t = xstransact(path='/local/domain/')
    domids = t.list()
    for domid in domids:
        path = t.read('%s/vm' % domid)
        if path == vm_path:
            domid = int(domid)
            log.debug("Domain %s has ID %d" % (uuid, domid))
            return domid
    log.debug("Domain %s is not running" % uuid)
    return None

def wait_for_uuid_to_exit(uuid):
    wait_for_domid_to_exit(get_domid_from_uuid(uuid), uuid)

def wait_for_domid_to_exit(domid, uuid):
    """Wait for domid to exit (or, for domid's UUID to not match
    the specified uuid, which would indicate a race condition in
    which domid has already exited and was replace by another)."""
    if domid is None:
        return
    uuid_path = '/local/domain/%d/vm' % domid
    log.debug("Watching xenstore path '%s'" % uuid_path)
    cv = threading.Condition()
    def fn(path):
        log.debug("Watch fired for '%s'" % uuid_path)
        cv.acquire()
        u = xstransact.Read(uuid_path)
        log.debug("UUID is '%s'", u)
        if u is None:
            again = False
        else:
            u = u.split('/')[-1]  # Why does uuid have "/vm/" prefix?
            if u == uuid:
                again = True
            else:
                again = False
        if not again:
            cv.notifyAll()
        cv.release()
        return again
    cv.acquire()
    w = xswatch(uuid_path, fn)
    cv.wait()
    cv.release()

def get_vnc_port_from_domid(domid, timeout=0.0):
    if domid is None:
        return None
    ports = [None]
    console_path = '/local/domain/%d/console/vnc-port' % domid
    log.debug("Watching xenstore path '%s'" % console_path)
    cv = threading.Condition()
    def fn(path):
        log.debug("Watch fired for '%s'" % console_path)
        again = True
        cv.acquire()
        p = xstransact.Read(console_path)
        if p is not None:
            log.debug("vnc-port is '%s'", p)
            ports[0] = p
            cv.notifyAll()
            again = False
        cv.release()
        return again
    cv.acquire()
    w = xswatch(console_path, fn)
    cv.wait(timeout)
    cv.release()

    port = ports[0]
    log.info('Domain ID %d has vnc port %s' % (domid, port))
    if port is not None:
        port = int(port)
    return port

def get_tty_from_domid(domid):
    if domid is None:
        return None
    console_path = '/local/domain/%d/console/tty' % domid
    tty = xstransact.Read(console_path)
    log.info('Domain ID %d has tty %s' % (domid, str(tty)))
    return tty

def vnc_console(guest, detach=False):
    """Forks a vncviewer process to view the domain.  Returns the
    child's PID if (success and not detach), else None.  Note that
    failure cannot be determined if detach is True.
    """
    domid = get_domid_from_uuid(guest.uuid)
    vncport = get_vnc_port_from_domid(domid, timeout=10.0)
    if vncport == None:
        log.info("Did not find VNC port, so not using VNC.")
        raise NotImplementedError
    if not os.environ.has_key('DISPLAY'):
        raise RuntimeError(msg.vnc_no_display % ('localhost', vncport))
    def find_viewer():
        try:
            for path in os.environ['PATH'].split(':'):
                viewer = os.path.join(path, 'vncviewer')
                if os.path.exists(viewer):
                    log.info("Found vncviewer: %s" % (viewer,))
                    return viewer
        except:
            pass
        return None
    def exec_viewer(viewer):
        try:
            os.execv(viewer, ['vncviewer', '::%d' % vncport, ])
        except:
            os._exit(1)
    viewer = find_viewer()
    if not viewer:
        raise RuntimeError(msg.display_failed)
    if not detach:
        pid = os.fork()
        if pid == 0:
            util.close_fds()
            exec_viewer(viewer)
    else:
        if util.daemonize():
            exec_viewer(viewer)
        pid = None
    log.info("Opened a VNC viewer to %s:%d." % ('localhost', vncport))
    return pid

def text_console(guest, detach=False):
    """Forks a text console process to view the domain.  Returns the
    child's PID if (success and not detach), else None.  Note that
    failure cannot be determined if detach is True.
    """
    domid = get_domid_from_uuid(guest.uuid)
    tty = get_tty_from_domid(domid)
    if tty is None or not os.access(tty, os.R_OK | os.W_OK):
        raise NotImplementedError
    if os.environ.has_key('DISPLAY'):
        vmdefaults = VMDefaults[guest.options.os_type]
        cmd = '/usr/bin/xterm'
        args = ['/usr/bin/xterm']
        args.extend(vmdefaults.xterm_args())
        args.extend(['-T', guest.options.name, '-e', 'xm', 'console', guest.uuid])
        log.info("Opened a console in an xterm.")
    else:
        cmd = '/usr/sbin/xm'
        args = ['/usr/sbin/xm', 'console', guest.uuid ]
    log.debug("Running: %s" % str(args))
    if not detach:
        pid = os.fork()
        if pid == 0:
            os.execvp(cmd, args)
            os._exit(1)
    else:
        if util.daemonize():
            os.execvp(cmd, args)
            os._exit(1)
        pid = None
    return pid

def reset():
    global _reset_console
    if _reset_console:
        os.system('reset')
        _reset_console = False
