# Exceptions
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


# Do not change self.base or self.err numbering in a way that breaks
# backwards compatibility!

import msg

class Error(Exception):
    """Base class for exceptions that occur when installing Xen VMs.

    Attributes:
        err -- exception-class-specific error code
        title -- human-readable generic description of 'err'
        details -- string
    """
    def __init__(self, err, title="Error", details=None):
        Exception.__init__(self, title)
        self.err = err
        self.title = title
        self.details = details

    def __str__(self):
        s = self.title
        if self.details:
            s += ' (%s)' % self.details
        return s

    def exit_code(self):
        return self.base + self.err

class XenError(Error):
    """Raised when an error occurs in Xen or while interacting with Xen."""
    base = 10
    titles = (
        msg.exc_xen_err,
        msg.exc_xen_none,
        msg.exc_xen_ver,
        msg.exc_xen_xend,
        msg.exc_xen_root)
    (E_OTHER,
     E_NO_XEN,
     E_VERSION,
     E_CONNECT,
     E_ROOT) = range(5)

    def __init__(self, err, details=None):
        Error.__init__(self, err, title=self.titles[err], details=details)

class VmError(Error):
    """Raised when an error occurs in regarding a VM."""
    base = 20
    titles = (
        msg.exc_vm_start,
        msg.exc_vm_crash,
        msg.exc_vm_exists,
        msg.exc_vm_conf_exists,
        msg.exc_vm_bootloader,
        msg.exc_vm_kernel)
    (E_NO_START,
     E_CRASH,
     E_VM_EXISTS,
     E_CONF_EXISTS,
     E_NO_BOOTLOADER,
     E_NO_KERNEL) = range(6)

    def __init__(self, err, details=None):
        Error.__init__(self, err, title=self.titles[err], details=details)

class InstSrcError(Error):
    """Raised when an error occurs with the installation source.
    """
    base = 30

    titles = (
        msg.exc_src_unsup,
        msg.exc_src_ioerr,
        msg.exc_src_no,
        msg.exc_src_bad,
        msg.exc_src_undef,
        msg.exc_src_cdrom)

    (E_UNSUPPORTED,
     E_IO,
     E_NOT_FOUND,
     E_BAD,
     E_UNDEF,
     E_CDROM) = range(6)

    def __init__(self, err=None, urlgrabber=None, details=None):
        if urlgrabber:
            errors = {
                1: InstSrcError.E_UNSUPPORTED, # malformed url
                2: InstSrcError.E_NOT_FOUND,   # local file doesn't exist
                3: InstSrcError.E_BAD,         # non-file local file (dir, etc)
                4: InstSrcError.E_IO,          # IOError
                5: InstSrcError.E_IO,          # OSError
                6: InstSrcError.E_IO,          # no content-length
                7: InstSrcError.E_IO,          # HTTPException
                8: InstSrcError.E_IO,          # Exceeded read limit (for urlread)
                9: InstSrcError.E_IO,          # Requested byte range not satisfiable.
                10:InstSrcError.E_IO,          # Byte range requested, but range support unavailable
                11:InstSrcError.E_IO,          # Illegal reget mode
                12:InstSrcError.E_IO,          # Socket timeout
                13:InstSrcError.E_UNSUPPORTED, # malformed proxy url
                14:InstSrcError.E_IO,          # HTTPError (includes .code and .exception attributes)
                15:InstSrcError.E_IO,          # user abort
            }
            err = errors[urlgrabber]
        if err is None:
            err = InstSrcError.E_BAD
        Error.__init__(self, err, title=self.titles[err], details=details)

class ResourceError(Error):
    def __init__(self, err, title, details=None):
        Error.__init__(self, err, title=title, details=details)

class ArchResourceError(ResourceError):
    base = 40

    titles = (
        msg.exc_arch_incompat,
        msg.exc_arch_novt)

    (E_INCOMPAT_ARCH,
     E_NO_VT) = range(2)

    def __init__(self, err, details=None):
        ResourceError.__init__(self, err, title=self.titles[err], details=details)

class MemoryResourceError(ResourceError):
    base = 50

    titles = (msg.exc_mem_no, )

    (E_NO_MEMORY,) = range(1)

    def __init__(self, err, details=None):
        ResourceError.__init__(self, err, title=self.titles[err], details=details)

class DiskResourceError(ResourceError):
    base = 60

    titles = (
        msg.exc_disk_no,
        msg.exc_disk_ro,
        msg.exc_disk_perm,
        msg.exc_disk_conflict)

    (E_NOT_FOUND,
     E_READ_ONLY,
     E_PERM,
     E_CONFLICT) = range(4)

    def __init__(self, err, details=None):
        ResourceError.__init__(self, err, title=self.titles[err], details=details)

class JobError(Error):
    base = 70

    titles = (
        msg.exc_job_canceled,
        msg.exc_job_no,
        msg.exc_job_state)
    (E_CANCELED,
     E_NOT_FOUND,
     E_JOB_STATE) = range(3)

    def __init__(self, err, details=None):
        Error.__init__(self, err, title=self.titles[err], details=details)

class ParamError(Error):
    base = 80

    titles = (
        msg.exc_param_invalid,
        msg.exc_param_nopara,
        msg.exc_param_nofull,
        msg.exc_param_noauto,
    )

    (E_INVALID,
     E_NOPARA,
     E_NOFULL,
     E_NOAUTO,
    ) = range(4)

    def __init__(self, err, details=None):
        Error.__init__(self, err, title=self.titles[err], details=details)
