# Network Interface Card class
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import string
import random

from exceptions import *
import msg

bridges = None
def get_bridges():
    global bridges
    if bridges is None:
        bridges = []
        rfd = os.popen("brctl show")
        data = rfd.read()
        status = rfd.close()
        lines = data.splitlines()[1:]
        for line in lines:
            if line[0] in (' ', '\t'):
                continue
            words = line.split()
            bridges.append(words[0])
    return bridges

def parse_string(nic):
    if nic == "":
        return XenNic()
    params = nic.split(',')
    mac = None
    bridge = None
    model = None
    for p in params:
        if p.startswith('mac='):
            mac = p[4:]
        elif p.startswith('bridge='):
            bridge = p[7:]
        elif p.startswith('model='):
            model = p[6:]
        else:
            raise ParamError(err=ParamError.E_INVALID, details=p)
    return XenNic(mac=mac, bridge=bridge, model=model)

class XenNic(object):

    models = {
        'pcnet': msg.nic_pcnet,
        'ne2k_isa': msg.nic_ne2k_isa,
        'ne2k_pci': msg.nic_ne2k_pci,
        'rtl8139': msg.nic_rtl8139,
        'para': msg.paravirtualized
    }

    def __init__(self, number=None, mac=None, bridge=None, model=None, full_virt=True):
        self.number = number       # free-form, for caller's use
        self.mac = mac             # None or '11:22:33:44:55:66'
        self.bridge = bridge
        self.full_virt = full_virt
        self.model = model         # None or one of XenNic.models.keys()

    def get_mac(self):
        return self._mac
    def set_mac(self, mac):
        def badmac():
            raise ParamError(err=ParamError.E_INVALID, details=msg.mac_address)
        if mac is not None and type(mac) != str:
            badmac()
        if mac:
            octets = mac.split(':')
            if len(octets) != 6:
                badmac()
            for o in octets:
                if len(o) != 2 or o[0] not in string.hexdigits or o[1] not in string.hexdigits:
                    badmac()
        if mac:
            self._mac = mac.lower()
        else:
            self._mac = None
    mac = property(get_mac, set_mac)

    def get_bridge(self):
        return self._bridge
    def set_bridge(self, bridge):
        self._bridge = bridge
    bridge = property(get_bridge, set_bridge)
    
    def get_model(self):
        if self._model is None:
            if self.full_virt:
                return 'rtl8139'
            else:
                return 'para'
        return self._model
    def set_model(self, model):
        if model is not None and model not in XenNic.models.keys():
            raise ParamError(err=ParamError.E_INVALID, details='model')
        self._model = model
    model = property(get_model, set_model)

    def copy(self):
        return XenNic(number=self.number, mac=self.mac, bridge=self.bridge, model=self._model, full_virt=self.full_virt)

    def setup(self, progress=lambda f:None):
        if self.mac is None:
            # Courtesy of netif.py randomMAC() to avoid circular imports
            mac = [ 0x00, 0x16, 0x3e, 
                    random.randint(0x00, 0x7f), random.randint(0x00, 0xff), random.randint(0x00, 0xff) ]
            self.mac = ':'.join(map(lambda x: "%02x" % x, mac))

        progress(1.0)

    def update(self, full_virt):
        if self._model is not None and self._model != 'para' and not full_virt:
            raise ParamError(err=ParamError.E_INVALID, details=str(self))
        self.full_virt = full_virt

    def __str__(self):
        s = '%s; ' % self.models[self.model]
        if self.mac is None:
            s += msg.random_mac
        else:
            s += self.mac
        return s

    def __repr__(self):
        r = ''
        if self.number is not None:
            r += '%d:' % self.number
        mac = self.mac or ''
        r += '%s,%s,%s,%d' % (self.mac or '', self.bridge or '', self.model or '', self.full_virt)
        return r
