# Options scratch-area for VM settings
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import string

from VMDefaults import VMDefaults, default_os_type
from exceptions import *
from log import log
from nics import XenNic
import caps
import msg
import util
import vmdisks
import xend

class Options(dict):
    def __init__(self):
        default_keys = (
            'apic',
            'background',
            'debug',
            'disks',            # Disk class
            'extra_args',       # User's install or boot args
            'full_virt',
            'graphics',         # 'cirrus', 'none', 'para', 'vesa'
            'graphics_viewer',  # 'text', 'sdl', 'vnc'
            'initrd',
            '_install_args',    # install-time args calculated by OS
            'install',
            'kernel',
            'loader',
            'max_memoryMB',
            'memoryMB',
            'name',
            'nics',             # XenNic class
            'noautoconsole',
            'norestart',
            'on_crash',
            'on_poweroff',
            'on_reboot',
            'os_settings',
            'os_type',
            '_runtime_args',    # runtime-time args calculated by OS
            'source',
            'ui',
            'usb',
            'usb_mouse_abs',
            'uuid',
            'vcpus',
            'vnc_password',
            'vncport',
            'pxe_boot',
        )
        dict.__init__(self, dict(map(lambda k: (k, None), default_keys)))
        self.__initialized = True

    def copy(self):
        dst = Options()
        for (key, value) in self.__dict__.items():
            if key == 'disks' and value is not None:
                dst.disks = []
                for disk in value:
                    dst.disks.append(disk.copy())
            elif key == 'nics' and value is not None:
                dst.nics = []
                for nic in value:
                    dst.nics.append(nic.copy())
            else:
                dst.__dict__[key] = value
        return dst

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError(item)

    def __setattr__(self, item, value):
        if not self.__dict__.has_key('_Options__initialized'):
            return dict.__setattr__(self, item, value)
        self.__setitem__(item, value)

    def __setitem__(self, item, value):
        """Sets an option, doing only basic order-independent validation
        such as type- and platform-validation."""

        log.debug("Setting options[%s] = %s" % (item, str(value)))

        # Raise AttributeError if attempting to add additional keys
        self.__getattr__(item)

        if value is None:
            self.__dict__[item] = None
            return

        def bad():
            raise ParamError(err=ParamError.E_INVALID, details=item)
        def check_type(v, types):
            if not type(v) in types:
                bad()

        if item == 'apic':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'ui':
            if value not in ('gtk', 'text', 'yast'):
                bad()
        elif item == 'background':
            check_type(value, (bool,))
        elif item == 'debug':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'disks':
            check_type(value, (list,))
            disks = value
        elif item == 'full_virt':
            check_type(value, (bool,))
            if value and not caps.is_hvm_capable():
                raise ArchResourceError(err=ArchResourceError.E_NO_VT)
        elif item == 'graphics':
            if value not in ('cirrus', 'none', 'para', 'vesa'):
                bad()
        elif item == 'install':
            check_type(value, (bool,))
        elif item == 'loader':
            if not os.path.exists(value):
                bad()
        elif item == 'max_memoryMB':
            check_type(value, (int, float, str))
            value = int(value)
        elif item == 'memoryMB':
            check_type(value, (int, float, str))
            value = int(value)
        elif item == 'name':
            check_type(value, (str,))
            validchars = string.ascii_letters + string.digits + '_-.:+'
            for c in value:
                if not c in validchars:
                    raise ParamError(err=ParamError.E_INVALID, details=msg.name_alpha)
            if len(value) == 0 or len(value) > 255:
                raise ParamError(err=ParamError.E_INVALID, details=msg.name_len)
            if value[0] in string.digits:
                raise ParamError(err=ParamError.E_INVALID, details=msg.name_start)
            if xend.connection.domain(value):
                raise VmError(err=VmError.E_CONF_EXISTS)
        elif item == 'nics':
            check_type(value, (list,))
            nics = value
        elif item == 'on_crash' or item == 'on_poweroff' or item == 'on_reboot':
            if value not in ('restart', 'destroy', 'preserve', 'rename-restart'):
                bad()
        elif item == 'os_settings':
            check_type(value, (str,))
        elif item == 'os_type':
            if not value in VMDefaults:
                bad()
        elif item == 'source':
            # str:  'ftp://foo/bar', 'dev:/xvda', etc
            # Disk:  source follows disk, even if disk gets renamed, unless
            #           somehow the disk becomes unbootable (e.g., second CD in FV)
            check_type(value, (str, vmdisks.Disk))
        elif item == 'usb':
            check_type(value, (bool,))
        elif item == 'usb_mouse_abs':
            check_type(value, (bool,))
        elif item == 'uuid':
            if type(value) == str:
                # FIXME:  Check length, etc.
                pass
            elif type(value) == type(int):
                value = util.uuidToString(value)
            else:
                bad()
        elif item == 'vcpus':
            check_type(value, (int, float))
            value = int(value)
        if item == 'pxe_boot':
            check_type(value, (int, bool))
            value = bool(value)

        self.__dict__[item] = value

    def para_possible(self):
        vmdefaults = VMDefaults[self.os_type]
        return 'xen' in vmdefaults.virt_support()

    def full_possible(self):
        vmdefaults = VMDefaults[self.os_type]
        return caps.is_hvm_capable() and ('hvm' in vmdefaults.virt_support() or
                                          'hvm-ext' in vmdefaults.virt_support())

    def full_preferred(self, user_request_full):
        vmdefaults = VMDefaults[self.os_type]
        supported = vmdefaults.virt_support()
        if self.para_possible():
            if self.full_possible():
                if user_request_full is not None:
                    return user_request_full
                return supported[0] in ('hvm', 'hvm-ext')
            else:
                if user_request_full:
                    if ('hvm' in supported or 'hvm-ext' in supported):
                        raise ArchResourceError(err=ArchResourceError.E_NO_VT)
                    else:
                        raise ParamError(err=ParamError.E_NOFULL)
                return False
        else:
            if self.full_possible():
                return True
            else:
                if ('hvm' in supported or 'hvm-ext' in supported):
                    raise ArchResourceError(err=ArchResourceError.E_NO_VT)
                else:
                    raise ParamError(err=ParamError.E_NOPARA)
            return True

def shallow_copy(src, dst):
    for key in src.__dict__.keys():
        value = getattr(src, key)
        dst.__dict__[key] = value

def calculate(options, defaults):
    """Re-calculates 'defaults' based on 'options'.  Error if 'options' is
    not self-consistent."""
    log.debug("start recalculation")

    if options.os_type:
        defaults.os_type = options.os_type
    else:
        defaults.os_type = default_os_type
    vmdefaults = VMDefaults[defaults.os_type]

    if options.pxe_boot:
        log.debug("using pxe boot, defaulting to full virtualization")
        defaults.pxe_boot = options.pxe_boot
        defaults.full_virt = True
    else:
        defaults.pxe_boot = False
        defaults.full_virt = defaults.full_preferred(options.full_virt)

    if options.name:
        defaults.name = options.name
    else:
        i = 0
        while True:
            name = defaults.os_type
            if i:
                name += '-%d' % i
            try:
                defaults.name = name
                break
            except VmError:
                i += 1
            except:
                defaults.name = None
                break

    defaults.apic = vmdefaults.apic()

    if options.memoryMB:
        defaults.memoryMB = options.memoryMB
    else:
        defaults.memoryMB = vmdefaults.default_memory()
    if options.max_memoryMB:
        defaults.max_memoryMB = options.max_memoryMB
    else:
        defaults.max_memoryMB = vmdefaults.max_memory()
    if defaults.memoryMB < vmdefaults.min_memory():
        # For now, don't go below OS's min, even if it's not available.
        # Availability will be checked later.
        defaults.memoryMB = vmdefaults.min_memory()
    if defaults.max_memoryMB < defaults.memoryMB:
        defaults.max_memoryMB = defaults.memoryMB

    if options.vcpus:
        defaults.vcpus = options.vcpus
    else:
        defaults.vcpus = vmdefaults.num_vcpus()

    if options.loader:
        defaults.loader = options.loader
    else:
        if defaults.full_virt:
            defaults.loader = '/usr/lib/xen/boot/hvmloader'
        else:
            defaults.loader = '/usr/lib/xen/boot/domUloader.py'

    # The command line can take a UUID, but the TUI or GUI won't let the
    # user specify one.  So, if one didn't exist on the command line,
    # stick with the first randomly generated one.
    if options.uuid is None:
        options.uuid = util.uuidToString(util.randomUUID())
    defaults.uuid = options.uuid

    if options.on_crash:
        defaults.on_crash = options.on_crash
    else:
        defaults.on_crash = 'destroy'

    if options.on_poweroff:
        defaults.on_poweroff = options.on_poweroff
    else:
        defaults.on_poweroff = 'destroy'

    if options.on_reboot:
        defaults.on_reboot = options.on_reboot
    else:
        defaults.on_reboot = 'restart'

    if options.graphics_viewer == 'sdl':
        defaults.graphics_viewer = 'sdl'
    else:
        defaults.graphics_viewer = 'vnc'
        defaults.vnc_password = options.vnc_password or ''
        # Nasty hack due to the fact that xen's config file has no escaping
        # FIXME:  drop support for xen's config file; move to XenAPI or libvirt
        if "'" in defaults.vnc_password or "," in defaults.vnc_password:
            raise ParamError(err=ParamError.E_INVALID, details='vnc_password')
        defaults.vncport = options.vncport

    if options.graphics == 'none':
        defaults.graphics = 'none'
    elif defaults.full_virt:
        if options.graphics in ('cirrus', 'vesa'):
            defaults.graphics = options.graphics
        else:
            defaults.graphics = 'cirrus'
    else:
        defaults.graphics = 'para'

    if options.install is None:
       defaults.install = True
    else:
       defaults.install = options.install

    if options.disks is not None:
        defaults.disks = options.disks
    else:
        if defaults.install:
            defaults.disks = vmdefaults.disks()
        else:
            defaults.disks = []  # Not installing; user must specify existing disk
    for d in defaults.disks:
        d.update(full_virt=defaults.full_virt, vmname=defaults.name)
        d.disk_group = defaults.disks
    if len(defaults.disks):
        disks = defaults.disks
        log.debug('Checking for conflicts between disks: %s' % map(str, disks))
        for i in range(len(disks)-1):
            for j in range(i+1, len(disks)):
                if (disks[i].vdev.startswith(disks[j].vdev) or
                    disks[j].vdev.startswith(disks[i].vdev) or
                    disks[i].pdev == disks[j].pdev):
                    raise DiskResourceError(err=DiskResourceError.E_CONFLICT,
                                            details='%s\n%s' % (disks[i], disks[j]))
    defaults.disks.sort(lambda d1,d2: cmp(d1.vdev, d2.vdev))

    if options.nics is not None:
        defaults.nics = options.nics
    else:
        defaults.nics = vmdefaults.nics()
    nics = defaults.nics
    if len(nics):
        for i in range(len(nics)-1):
            for j in range(i+1, len(nics)):
                if (nics[i].mac is not None and
                    nics[i].mac == nics[j].mac):
                    raise ParamError(err=ParamError.E_INVALID, details=msg.mac_address)
    for n in nics:
        n.update(defaults.full_virt)

    if options.extra_args:
        defaults.extra_args = options.extra_args
    else:
        defaults.extra_args = ''

    defaults.source = None
    if options.source:
        defaults.source = options.source
        # FIXME:  full vs para affects what protocols are supported!
        source = defaults.source
        if type(source) is vmdisks.Disk:
            source = 'dev:/'
        supported = False
        protocols = vmdefaults.install_sources()[1]
        for protocol in protocols:
            if source.startswith(protocol):
                supported = True
                break
        if not supported:
            raise InstSrcError(err=InstSrcError.E_UNSUPPORTED, details=','.join(protocols))
        if type(defaults.source) is str and defaults.source.startswith('dev:/'):
            # User specified string; convert to disk.
            vdev = defaults.source[5:]
            for disk in defaults.disks:
                if str(disk.get_vdev_index()) == vdev or disk.get_vdev_str() == vdev:
                    defaults.source = disk
                    break
        if type(defaults.source) is vmdisks.Disk and defaults.source not in defaults.disks:
            log.debug("source disk has been removed; re-defaulting source")
            options.source = defaults.source = None
        if defaults.full_virt and type(defaults.source) is vmdisks.Disk and \
            defaults.source not in vmdisks.get_fv_boot_disks(defaults.disks):
            log.debug("source disk not valid to boot from for FV; re-defaulting source")
            options.source = defaults.source = None
    if defaults.source is None:
        if defaults.full_virt:
            fv_boot_disks = vmdisks.get_fv_boot_disks(defaults.disks)
            log.debug("possible FV boot disks: %s" % map(str, fv_boot_disks))
            if fv_boot_disks:
                defaults.source = fv_boot_disks[0]
        elif 'dev:/' in vmdefaults.install_sources()[1]:
            vdevs = filter(lambda d: d.exists(), defaults.disks)
            if vdevs:
                # Slight hack, to prefer cdroms over disks and floppies.
                vdevs.sort(lambda x,y: cmp(x._vdevType, y._vdevType))
                defaults.source = vdevs[0]

    if options.os_settings:
        defaults.os_settings = options.os_settings
        if defaults.os_settings and vmdefaults.auto_install() is None:
            raise ParamError(err=ParamError.E_NOAUTO)
    else:
        defaults.os_settings = ''

    defaults.vncport = options.vncport
    defaults.norestart = options.norestart
    defaults.debug = options.debug
    defaults.ui = options.ui
    defaults.background = options.background

    if defaults.background:
        defaults.noautoconsole = True
    else:
        defaults.noautoconsole = options.noautoconsole

    # User can't specify _install_args or _runtime_args,
    # but the OS-specific code may have.
    defaults._install_args = vmdefaults.actions().getInstallArgs(defaults)
    if options._install_args:
        defaults._install_args += ' %s' % options._install_args
    defaults._runtime_args = vmdefaults.actions().getRuntimeArgs(defaults)
    if options._runtime_args:
        defaults._runtime_args += ' %s' % options._runtime_args
    
    # Now give the OS a chance to voice its opinion:
    vmdefaults.actions().preprocessOptions(defaults)

    log.debug("end recalculation")

def finalize(options):
    """Checks the options for conflicts or incomplete options."""

    # The Options class will guarentee that no single setting is invalid,
    # but since it has no concept of ordering or completeness, it can't
    # check for missing values, conflicts, etc.

    vmdefaults = VMDefaults[options.os_type]

    if options.name is None:
        log.debug("validation failed: options.name is None")
        raise ParamError(err=ParamError.E_INVALID, details=msg.name_len)

    if len(options.disks) == 0:
        raise DiskResourceError(err=DiskResourceError.E_NOT_FOUND)

    for disk in options.disks:
        disk.validate()

    for nic in options.nics:
        if not options.full_virt and nic.model != 'para':
            raise ParamError(err=ParamError.E_INVALID, details=eval('msg.nic_' % nic.model))
        nic.setup()

    try:
        xen_info = xend.connection.info()
        if options.full_virt:
            avail_mb = xen_info['max_hvm_memory']
        else:
            avail_mb = xen_info['max_para_memory']
    except:
        avail_mb = options.memory
    if options.memoryMB > avail_mb:
        raise MemoryResourceError(err=MemoryResourceError.E_NO_MEMORY)

    if options.full_virt:
        if options.install:
            cdrom = None
            for disk in options.disks:
                if disk._vdevType == vmdisks.Disk.DEVICE_CDROM:
                    cdrom = disk
                    break
            if cdrom is None and options.pxe_boot is False:
                log.debug("validation failed: No CDROM found for FV VM")
                raise InstSrcError(err=InstSrcError.E_CDROM)

        if options.graphics not in ('cirrus', 'none', 'vesa'):
            raise ParamError(err=ParamError.E_INVALID, details='graphics')

        # Work-around for Qemu slowness (bug #238458):  Virtual CD-ROMs
        # slow to a crawl if they share the virtual IDE controller with
        # a hard disk.  Attempt to push the CD-ROM(s) to their own
        # controller, without changing the ordering, if possible.
        if (len(filter(lambda d: d.vdevtype == vmdisks.Disk.DEVICE_CDROM, options.disks)) > 0):
            if len(options.disks) == 2:
                (options.disks[0].vdev, options.disks[1].vdev) = (0, 2)
            elif len(options.disks) == 3:
                if map(lambda d: d.vdevtype, options.disks).index(vmdisks.Disk.DEVICE_CDROM) == 2:
                    (options.disks[0].vdev, options.disks[1].vdev, options.disks[2].vdev) = (0, 1, 2)
                else:
                    (options.disks[0].vdev, options.disks[1].vdev, options.disks[2].vdev) = (0, 2, 3)
            log.debug('Re-distributed disks: %s' % map(str, options.disks))

    else:
        if options.install:
            if not options.source:
                log.debug("validation failed: No install source found for PV VM")
                raise InstSrcError(err=InstSrcError.E_UNDEF)

            if type(options.source) is vmdisks.Disk:
                vdev = options.source.vdev
                if len(filter(lambda d: d.vdev==vdev and d.exists(), options.disks)) == 0:
                    log.debug("validation failed: No matching disk found for PV install source")
                    raise InstSrcError(InstSrcError.E_NOT_FOUND)

        if options.graphics not in ('none', 'para'):
            raise ParamError(err=ParamError.E_INVALID, details='graphics')
